/*
 * linux/drivers/ide/arm/ide-ep93xx.c
 * Copyright (C) 2007 Atmark Techno, Inc
 *
 * Adapted from original the ide-ep93xx.c, which is:
 *   Copyright (C) 2005 Cirrus Logic
 *
 * and which in turn was adapted from ide-io.c and also pmac.c, which is:
 *   Copyright (C) 1998-2003 Paul Mackerras & Ben. Herrenschmidt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 */

#include <linux/ide.h>
#include <asm/io.h>
#include <asm/ide.h>
#include <asm/arch/ide.h>
#include <asm/hardware.h>

#ifdef CONFIG_BLK_DEV_IDE_DMA_EP93XX
extern void ep93xx_ide_dma_init(ide_hwif_t *hwif);
#endif

/* 
 * one nop usually equals one cpu cycle (?), 
 * so one nop equals 5nsec with a 200MHz clock.
 */
#define nop5()	{ nop(); nop(); nop(); nop(); nop(); }
#define nop10()	{ nop5(); nop5(); }

/*
 * ep93xx_ide_outw - write out 16bits to specified ATA register address
 */
static void ep93xx_ide_outw(u16 w, unsigned long addr)
{
	unsigned int val;

	outl(w, IDEDATAOUT);

	/* select ATA register */
	val = IDECtrl_DIORn | IDECtrl_DIOWn | ((addr & 7) << 2) | (addr >> 10);
	outl(val, IDECR);

	nop5();

	/* toggle the write signal */

	outl(val & ~IDECtrl_DIOWn, IDECR);

	nop10(); nop(); nop(); nop(); nop();

	outl(val, IDECR);

	nop5();
}

/*
 * ep93xx_ide_inw - read 16bits from specified ATA register address
 */
static u16 ep93xx_ide_inw(unsigned long addr)
{
	unsigned int val;

	/* select ATA register */
	val = IDECtrl_DIORn | IDECtrl_DIOWn | ((addr & 7) << 2) | (addr >> 10);
	outl(val, IDECR);

	nop5();

	/* toggle the read signal */

	outl(val & ~IDECtrl_DIORn, IDECR);

	nop10(); nop(); nop(); nop(); nop();

	outl(val, IDECR);

	return (inl(IDEDATAIN) & 0xffff);
}

/*
 * ep93xx_ide_outb - write out one byte to specified ATA register address
 */
static void ep93xx_ide_outb(u8 b, unsigned long addr)
{
	ep93xx_ide_outw(b, addr);
}

static void ep93xx_ide_outbsync(ide_drive_t *drive, u8 b, unsigned long addr)
{
	ep93xx_ide_outw(b, addr);
}

/*
 * ep93xx_ide_inb - read one byte from specified ATA register address
 */
static unsigned char ep93xx_ide_inb(unsigned long addr)
{
	return ep93xx_ide_inw(addr);
}

static void ep93xx_ide_insw(unsigned long addr, void *buf, u32 count)
{
	unsigned short *data = (unsigned short *)buf;

	for (; count; count--)
		*data++ = ep93xx_ide_inw(addr);
}

static void ep93xx_ide_outsw(unsigned long addr, void *buf, u32 count)
{
	unsigned short *data = (unsigned short *)buf;

	for (; count; count--)
		ep93xx_ide_outw(*data++, addr);
}

static void ep93xx_ata_input_data(ide_drive_t *drive, void *buffer, u32 count)
{
	ep93xx_ide_insw(IDE_DATA_REG, buffer, count << 1);
}

static void ep93xx_ata_output_data(ide_drive_t *drive, void *buffer, u32 count)
{
	ep93xx_ide_outsw(IDE_DATA_REG, buffer, count << 1);
}

static void ep93xx_atapi_input_bytes(ide_drive_t *drive,
	void *buffer, u32 count)
{
	ep93xx_ide_insw(IDE_DATA_REG, buffer, (count >> 1) + (count & 1));
}

static void ep93xx_atapi_output_bytes(ide_drive_t *drive, 
	void *buffer, u32 count)
{
	ep93xx_ide_outsw(IDE_DATA_REG, buffer, (count >> 1) + (count & 1));
}

void ep93xx_ide_init(ide_hwif_t *hwif)
{
	unsigned long value;

	/* make sure the GPIO on IDE bits in the DEVCFG register are not set. */
	value = inl(SYSCON_DEVCFG) & ~(SYSCON_DEVCFG_EonIDE |
				       SYSCON_DEVCFG_GonIDE |
				       SYSCON_DEVCFG_HonIDE);

	SysconSetLocked(SYSCON_DEVCFG, value);

	/* insure that neither device is selected */
	outl(inl(IDECR) | (IDECtrl_CS0n | IDECtrl_CS1n), IDECR);

	/* clear the MDMA and UDMA operation registers */
	outl(0, IDEMDMAOP);
	outl(0, IDEUDMAOP);

	/* enable PIO in mode 4 */
	outl(IDECfg_PIO | IDECfg_IDEEN | (4 << IDECfg_MODE_SHIFT) |
		(1 << IDECfg_WST_SHIFT), IDECFG);

	hwif->ata_input_data 		= ep93xx_ata_input_data;
	hwif->ata_output_data 		= ep93xx_ata_output_data;
	hwif->atapi_input_bytes 	= ep93xx_atapi_input_bytes;
	hwif->atapi_output_bytes 	= ep93xx_atapi_output_bytes;

	hwif->OUTB 	= ep93xx_ide_outb;
	hwif->OUTBSYNC 	= ep93xx_ide_outbsync;
	hwif->OUTW 	= ep93xx_ide_outw;
	hwif->OUTSW 	= ep93xx_ide_outsw;

	hwif->INB 	= ep93xx_ide_inb;
	hwif->INW 	= ep93xx_ide_inw;
	hwif->INSW 	= ep93xx_ide_insw;

#ifdef CONFIG_BLK_DEV_IDE_DMA_EP93XX
	ep93xx_ide_dma_init(hwif);
#endif
}

